#!/usr/bin/perl

# Copyright (c) 1994 by Vixie Enterprises
# 
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies, and that
# the name of Vixie Enterprises not be used in advertising or publicity
# pertaining to distribution of the document or software without specific,
# written prior permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND Vixie Enterprises DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL Vixie Enterprises
# BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
# OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

# $Id: do-hosts.pl,v 1.23 1995/05/11 12:51:29 brister Exp $

# Credits:
#       Paul Vixie <paul@vix.com> (functional design partner; original coder)
#       Jerry Scharf <scharf@sony.com> (functional design partner)
#       James Brister <brister@vix.com> (major code rework; maintainance)

#
# Process the host files in group areas. The output are DNS database files.
# 

#
# Boilerplate
#

$sends'Root = $ENV{'SENDS_ROOT'} ;
$sends'Root = "/usr/local/sends" unless defined ($sends'Root) ;
require "$sends'Root/lib/sends-config.pl" ;

##################################################

$DoNotEdit = "this file was automatically generated - do not edit";

@NoNoDomains = ( split (' ',$sends'Config{'bad_domains'}) , &sends'DomainList);

$HostOpt = $sends'zoneDir.'/hostoptions';
$HostDom = $sends'zoneDir.'/hostdomains';

#  $GlobalZoneMask = $sends'Config{'zonemask'} ;
#  $GlobalZoneMask = oct ($GlobalZoneMask) if $GlobalZoneMask =~ /^0/ ;

# $ZoneSize = $sends'Config{'zonesize'} ;
# $ZoneSize = oct ($ZoneSize) if $ZoneSize =~ /^0/ ;


%SupportedHostOptions = (
	'd', '',		# dumb -- smtp should strip its name
	'u', '',		# has unique local aliases
	'g', '',		# does not want/need an mx to itself
	'p', '',		# MXMail postoffice entry (blech).
	'', '' );


%ZoneMaskByAddr = () ;
%Control = ();
%FwdZones = ();		# by {$fqdn}
%RevZones = ();		# by {$1.$2.$3}
%BasenameByAddr = ();
%BasenameByAlias = ();
%AliasesByAddr = ();
%AddrsByAlias = ();
%AddrsByBasename = ();
%AliasesByBasename = ();
%OptionsByBasename = ();
%MailDomByBasename = ();
%MailDomsByZone = ();
%RelaysByMailDom = ();
%mxList = () ;
@MXDomains = () ;
%is1101 = () ;

$MissingOk = 1 ; # flag for some calls to get_input

$ExitStatus = 0;
&parse_options();		# get workdomain and workgroup

$opt_d = 1 if $sends'Config{'debug'} ;

$WorkFile = 0;

#
# Process ALL groups except for the one we've explicitly told to process
# ($WorkDomain, $WorkGroup).
#
&sends'forEachGroup("&main'doNonworkGroup()");

if ($opt_w) {
	local($fqdn) = &sends'fqdnByGroup($WorkDomain, $WorkGroup);

	if (!defined($fqdn)) {
		print STDERR "fqdnByGroup FAILED: $!\n";
		exit(1);
	}

	$WorkFile = 1;

	print STDERR "(new: $fqdn)\n"			if $opt_d;

	#
	# Get the control information (network, realy, dots-allowed), and
	# then process the group's input file.
	#
	&get_controls($WorkDomain, $WorkGroup);
	$_ = &get_input($fqdn, &sends'MailDomByGroup($WorkDomain,$WorkGroup),
			&sends'GroupToPath($WorkDomain,$WorkGroup)
				. '/work/hosts');
	die "$_\n" if $_;
}


@SortedAddrs = sort(keys(%BasenameByAddr));
&dump_it_out() if $opt_d;	# debugging
if ((!$opt_n) && (!$ExitStatus)) {
#	system("rm -f $sends'zoneDir/work/* >/dev/null 2>&1");
	foreach $zone (sort(keys(%FwdZones))) {
		&produce_fwddns($zone);
	}
	foreach $zone (sort(keys(%RevZones))) {
		&produce_revdns($zone);
	}
	&produce_hostdb();
}
exit($ExitStatus);

############################################################################
##
## doNonWorkGroup
##
## Called from inside the sends'forEachGroup routine.
## If the group we are given (via $sends'Domain and $sends'Group) is not
## the one that's being processed at the top level then we handle it here.
##
## We get the control information (network, relays, dots-allowed) for the
## group and then process the groups' input files.

sub doNonworkGroup {
	local($fqdn) = &sends'fqdnByGroup($sends'Domain, $sends'Group);
	local($_);

	if (!defined($fqdn)) {
		print STDERR "fqdnByGroup FAILED: $!\n";
		return 0;
	}

	#
	# make sure not the top level domain/group
	#
	return 1 if $opt_w
		&& ($sends'Domain eq $WorkDomain)
		&& ($sends'Group eq $WorkGroup);
	print STDERR "(nonWork: $fqdn)\n"			if $opt_d;

	&get_controls($sends'Domain, $sends'Group);
	
	# If the input file is missing that could be because this
	# is a new group that hasn't yet had a hosts file dropped in.
	$_ = &get_input($fqdn,
			&sends'MailDomByGroup($sends'Domain, $sends'Group),
			$sends'Path.'/output/hosts',
			$MissingOk);
	warn "$_\n" if $_;
	return 1;
}

############################################################################
##
## parse_options
##
## The usual.
##

sub parse_options {
	$opt_w = $opt_W = $opt_n = $opt_d = undef;
	&Getopts('w:W:fnd');
	if ((!$opt_w != !$opt_W) || (!$opt_f == !$opt_w)) {
		print STDERR
			"usage: $0 {-w wrkgrp -W wrkdom | -f} [-n] [-d]\n";
		exit(1);
	}
	if ($opt_w && $opt_W) {
		($WorkDomain, $WorkGroup) = ($opt_W, $opt_w);
	}
}

############################################################################
##
## get_controls
##
## Gets the 'relay' 'network' and 'dots-allowed' control variables for the 
## group. The relay is verified against the maildomain's relay. The network
## control value is verified for format and the dots-allowed is cached.
##

sub get_controls {
	local($domain, $group) = @_;
	local($maildom) = &sends'MailDomByGroup($domain, $group);
	local($_, $txt, $err, $relay, $mxrr);

	%Control = &sends'getControl($domain, $group);
	if (!scalar(%Control)) {
		print STDERR "$0: ".$sends'Error."\n";
		exit(1);
	}

	@MXDomains = () ;
	foreach $mxelem (split(/$;/, $Control{'mxdomains'})) {
		foreach $mx ( split(' ',$mxelem) ) {
#			if ($mx !~ /\.$/) {
#				print STDERR
#					"$0: mxdomains '$mx' is not a fqdn\n" ;
#				exit (1) ;
#			} else {
				push (@MXDomains,$mx) ;
#			}
		}
	}
			

	foreach $relay (split(/$;/, $Control{'relay'})) {
		foreach $mxrr (split(/\s+/, $relay)) {
			if ($relay !~ /\.$/) {
				print STDERR
					"$0: relay '$mxrr' is not a fqdn\n";
				exit(1);
			}
		}
	}

	#
	# if this maildomain has appeared before then the mail relays
	# better match
	#
	if (defined($RelaysByMailDom{$maildom})) {
		if ($RelaysByMailDom{$maildom} ne $Control{'relay'}) {
			print STDERR
				"$0: 'relay' mismatch for maildom $maildom\n";
 			print STDERR
				"\t(nonmatching group: [$domain,$group])\n";
			print STDERR
				"\tWARNING: ignoring nonmatching 'relay'\n";
		}
	} else {
		$RelaysByMailDom{$maildom} = $Control{'relay'};
	}

	@Networks = ();
	foreach (split(/$;/, $Control{'network'})) {
		($txt, $err) = &sends'parseNetworks($_);
		die "get_controls: $err\n" unless defined($txt);
		push(@Networks, $txt);
	}
	$DotsAllowed = ($Control{'dots-allowed'} =~ /[yY][eE][sS]/) ;
}


############################################################################
##
## Check that a hostname is legal.
##
## That is: no dots in name unless allowed and not in a disallowed domain.
##

sub checkHostName {
	local ($name,$fqdn,$dotsAllowed,$nonoDomains) = @_ ;
	local ($hname) ;

	if ($name =~ /\.$/) {
		return ('', "$name has a trailing dot");
	}

	$name =~ s/\.$fqdn//; # strip off trailing fqdn
	if (($name =~ /\./) && !$dotsAllowed) {
		return ('', "$name has a dot - not allowed") ;
	}

	foreach $domain (@nonoDomains) {
		# XXX - what about word boundaries?
		if ($name =~ /$domain$/) {
			return ('', "$name has unacceptable domain") ;
		}
	}
	$hname = join($;,$name,$fqdn) ;

	return ($hname,undef) ;
}

############################################################################
##
## genMXList
##
## Given a name, a fqdn and the rest of the special comment line, generate
## mx's. The comment line originally looks like:
##
##		#-mx-mailserv1:morality;opt1,hp0;hp1
##
## in this instance the routine will be called like:
##
##		genMXList ("mailserv1",$fqdn,"morality;opt1,hp0;hp1") 
##

sub genMXList {
	local ($name,$fqdn,$specs) = @_ ;
	local ($target) ;
	local (@mlist,$_,$err,$pri,$t,$tm) ;

##DEBUG
##	print STDERR "In genMXList (@_)\n" ;

	$pri = 0 ;
	while ( $specs =~ /^\s*([,\w\s.]+)\s*/ ) {
		@targets = split(',',$1) ;
		@targets = grep (s/\s*//g,@targets) ;
		($specs = $') =~ s/^;// ;
		foreach $tm ( @targets ) {
			$t = $tm ;
			# check that $t falls in the fqdn or the mxdomains.
			if ( $t =~ /\.$fqdn$/ ) {
				$t =~ s/\.$fqdn$// ;	
				$tm =~ s/\.$fqdn$// ;
			} else {
				foreach $m (@MXDomains) {
					if ($t =~ /\.$m/) {
						$t =~ s/\.$m// ;
						$tm .= "." ;
## DEBUG
##						print STDERR "MX DOMAIN found\n" ;
						last ;
					}
				}
			}

			if ($t =~ /\./ && !$DotsAllowed) {
				$err .= "MX Target $tm has dots--not valid. " ;
			}

			push (@mlist,$pri,$tm) ;
		}
	} continue {
		$pri++ ;
	}

	$mxList{$name} = join($;,@mlist) unless $err ;

##DEBUG
##	print STDERR "MX list for $name is\n\t", 
##			join("\n\t",split($;,$mxList{$name})), 
##			"\n" unless $err ;
##	print STDERR "Error was $err\n" if $err ;
	return $err ;
}


############################################################################
##
## get_input
##
## Process a near-standard format hosts file in the group's area.
## A slight format extension is that a host entry can have a comment at the
## end of the line that looks like:		# -dug-
## This entry is called a hostoption and is used to modify the standard
## processing steps for that host alone.
##
## We also now allow a full-line comment of the form
##	#-mx-mailserv1:morality;opt1,hp0;hp1
## this is used to generate MX records of the structure:
##	mailserv1	in mx	0 morality
##			in mx	1 opt1
##			in mx	1 hp0
##			in mx	2 hp1
##

sub get_input {
	local($fqdn, $maildom, $filename, $missingOk) = @_;
	local($options, $addr, $iaddr, @names, @aliases);
	local($basename, $name, $domain, $option);
	local($p, $error, $range, $action, $network, $mx);
	local(*f);

	print STDERR "get_input($fqdn, $maildom, $filename)\n" if $opt_d;
	if ( !open(f, $filename) ) {
		if ( $missingOk ) {
			print STDERR "Skipping missing file: $filename\n" if $opt_d ;
			return '' ;
		} else {
			print STDERR "Bad news, a missing file: $filename\n" if $opt_d ;
			return "$filename: $!";
		}
	}

	if ($maildom !~ /$fqdn$/) {
		close (f) ;
		return "maildom $maildom does not end in fqdn $fqdn";
	}
	if (defined($MailDomsByZone{$fqdn})) {
		$MailDomsByZone{$fqdn} = join($;, split(/$;/,
							$MailDomsByZone{$fqdn}
							),
						  $maildom);
	} else {
		$MailDomsByZone{$fqdn} = $maildom;
	}

line:	while (<f>) {
		chop;
		next if /^\s*$/ ;
		tr/A-Z/a-z/;

		next if ( /^#/ && substr ($_,0,5) ne "#-mx-" ) ;

		if (/^[\+\-]/) {
			&complain('NIS marker', 0);
			next line;
		}

		#
		# handle possible MX records and host options
		#  #-mx-mailserv1:morality;opt1,hp0;hp1
		#
		if (/^\#-mx-\s*(\w+)\s*:\s*/) {
			local ($rest) = $' ;
			local ($mx,$err) = &checkHostName($1,$fqdn,
							  $DotsAllowed,
							  @NoNoDomains) ;

			if ( !$mx ) {
				&complain($err,0) ;
				next line ;
			}

			$err = &genMXList ($mx,$fqdn,$rest) ;

			if ( $err ) {
				&complain ($err,0) ;
			}

			next line ;
		} elsif (/\#\-(\w*)\-/) {
			$options = $1;
			# convert from -du- to -d-u- form
			while ($options =~ /\-\w\w/) {
				$options =~ s/\-(\w)(\w)/\-$1\-$2/;
			}
			foreach $option (split(/\-/, $options)) {
				if (!defined($SupportedHostOptions{$option})) {
					&complain("bad option \"$option\"", 0);
					next line;
				}
			}
		} else {
			$options = '';
		}
		$_ = $` if /\#/;     # strip off host options and comments

		($addr, @names) = split;

		($f, $iaddr) = &sends'iaddrOf($addr); # turn into an integer
		if ($f) {
			&complain($f, 1);
			next line;
		}

		if ($#names < $[) {
			&complain('no host names found', 1);
			next line;
		}


		#
		# Now loop through all the networks that were defined
		# in the group's control file and make sure the host we're
		# processing falls within the legal address range.
		#
		$f = 'deny';
		$mask= '' ;
	net:	foreach $network (@Networks) {
			($range,$action,$netmask) = split(/\s+/, $network);
			if ($action eq 'deny' && $netmask) {
				die "Can't specifiy netmask and 'deny'\n" ;
			}
			($p, $error) = &sends'addrInRange($addr, $range);
			if (!defined($p)) {
				&complain($error, 1);
				next line;
			}
			if ($p) {
				$f = '' ;
				($f,$mask) = &sends'iaddrOf ($netmask) 
					if $netmask ;
				die "Processing netmask for $fqdn: $f\n" 
					if $f ;
				$f = $action ;
				last net;
			}
		}
		if ($f eq 'deny') {
			&complain("$addr is not in your network ACL", 0);
			next line;
		}

		# Find the RFC1101 mask for this address. It may have been
		# in the control file, or it will be in the global config file.
		($f,$RFCMask) = ($mask ? ('', $mask) : 
					&inetToRFC1101 ($iaddr)) ;
		if ((($iaddr & ~$RFCMask) == 0) && ($#names > 0)) {
			&complain ("networks cannot have aliases",1); 
			next line ;
		}

		#
		# Now loop over all the names that were in the host entry
		# and verify the hostname format:
		#	1 - can't be in an illegal domain (like junet)
		#	2 - can't have a '.' in the name unless the
		#	    dots-allowed control field was set.
		#	3 - can't be within any of the top-level domains
		#	    that we manage (e.g. hw.sony.com)
		#
		$f = 1;		# first one is special--the base name.
		$basename = undef;
		@aliases = ();
	name:	foreach $name (@names) {
			if (/\.$/) {
				&complain("$name has a trailing dot", $f);
				next name;
			}
			$name =~ s/\.$fqdn//; # strip off trailing fqdn
			if (($name =~ /\./) && !$DotsAllowed) {
				&complain("$name has a dot - not allowed", $f);
				next name;
			}
			foreach $domain (@NoNoDomains) {
				# XXX - what about word boundaries?
				if ($name =~ /$domain$/) {
					&complain(
					    "$name has unacceptable domain",
					    $f);
					next name;
				}
			}
			if ($f) {
				$basename = join($;, $name, $fqdn);
			} else {
				$p = join($;, $name, $fqdn);
				push(@aliases, $p)  unless $p eq $basename;
			}
		} continue {
			$f = 0;
		}
		if (!defined($basename)) {
			&complain("no valid basename found", 1);
			next line;
		}

		######################################## database validation
		#
		# Now we verify various things about all the names we found
		# on the host line:
		#	1 - The base name can't have already been used as
		# 	    an alias and vice-versa
		#	2 - If the basename is used more than once then
		#	    only one entry can have host options.
		# 	3 - Can't have two entries with the same iaddr.
		#	4 - An alias can't be an alias for two different
		#	    basenames 

		if (defined($BasenameByAddr{$iaddr})) {
			&complain("$addr used more than once", 1);
			next line;
		}

		if (defined($BasenameByAlias{$basename})) {
			&complain("$basename is already an alias", 1);
			next line;
		}

		foreach $name (@aliases) {
			if (defined($BasenameByAlias{$name})) {
				if ($BasenameByAlias{$name} ne $basename) {
					&complain(sprintf(
					    "%s is already an alias for %s",
					    $name, $BasenameByAlias{$name}),
					  1);
					next line;
				}
			}
			if (defined($AddrsByBasename{$name})
			  ||defined($MailDomByBasename{$name})
			   ) {
				&complain("$name is already a basename", 1);
				next line;
			}
		}

		if (defined($OptionsByBasename{$basename})) {
			if ($options ne $OptionsByBasename{$basename}) {
				&complain(sprintf(
				  "$basename can't have options '%s' AND '%s'",
					$OptionsByBasename{$basename},
					$options), 1);
				next line;
			}
		}

		########################################### database insertion
		#
		# The host information is good so now we stash it into the 
		# main arrays
		#

		$FwdZones{$fqdn} = '';

		# figure out which zone the address is in...
		($f,$ZoneMask) = &inetToZoneMask ($iaddr) ;
		die "$f\n" if $f ;

## DEBUG
		print "ZoneMask for ", &sends'fmtAddr ($iaddr), " = ", 
			&sends'fmtAddr ($ZoneMask), "\n" if $opt_d ;
		
		$f = '' ;
		($f, $RevZones{$iaddr & $ZoneMask}) = &inetToZoneSize ($iaddr) 
			unless defined ($RevZones{$iaddr & $ZoneMask}) ;
		die "$f\n" if $f ;
	
## DEBUG
		print "", &sends'fmtAddr ($iaddr), " is in zone ",
			join ('.',(split(/\./,&sends'fmtAddr 
				($iaddr & $ZoneMask)))
				[0..($RevZones{$iaddr & $ZoneMask} - 1)]),"\n"
			if $opt_d ;

		# save the zone mask 'cause we need it later when we're no
		# longer processing a specific group's input.
		$ZoneMaskByAddr{$iaddr} = $ZoneMask ;

		# check if it's a RFC1101 network address
		
		$is1101{$iaddr} = $RFCMask if (($iaddr & ~$RFCMask) == 0) ;

##DEBUG
		print "", &sends'fmtAddr ($iaddr), " is a NETNAME ",
			"according to the RFCMask ", &sends'fmtAddr($RFCMask),
			"\n" if (($iaddr & ~$RFCMask) == 0) && $opt_d ;


#		printf STDERR "Indexing zones %s and %s (%d %d)\n", 
#			$fqdn, &sends'fmtAddr ($iaddr & $ZoneMask), 
#			$iaddr, $ZoneMask if $opt_d ;

		# 	iaddr -> basename	(one-to-one)
		#	iaddr -> aliases	(can have multiple aliases)
		$BasenameByAddr{$iaddr} = $basename;
		if ($#aliases >= 0) {
			$AliasesByAddr{$iaddr} = join(' ', @aliases);
		}

		#	alias -> basename	(one-to-one)
		#	alias -> iaddr		(can have multiple iaddrs)
		foreach $name (@aliases) {
			$BasenameByAlias{$name} = $basename;
			if (defined($AddrsByAlias{$name})) {
				$AddrsByAlias{$name} =
					join($;,
					     split(/$;/, $AddrsByAlias{$name}),
					     $iaddr);
			} else {
				$AddrsByAlias{$name} = $iaddr;
			}
		}

		#
		# basename -> iaddr mapping	(can have multiple iaddrs)
		#
		if (defined($AddrsByBasename{$basename})) {
			$AddrsByBasename{$basename} =
				join($;,
				     split(/$;/, $AddrsByBasename{$basename}),
				     $iaddr);
		} else {
			$AddrsByBasename{$basename} = $iaddr;
		}

		#
		# basename -> alias mapping	(can have multiple aliases)
		# basename -> host_options	(one-to-one)
		# basename -> mail domain	(one-to-one)
		#
		if ($#aliases >= 0) {
		    if (defined($AliasesByBasename{$basename})) {
			$AliasesByBasename{$basename} =
				join(' ',
				    split(/\s/, $AliasesByBasename{$basename}),
				     @aliases);
		    } else {
			$AliasesByBasename{$basename} = join(' ', @aliases);
		    }
		}
		$OptionsByBasename{$basename} = $options;
		$MailDomByBasename{$basename} = $maildom
			unless $options =~ /\-u\-/;
	} #line
	close(f);

	#
	# Some sanity checks for maildomains and fqdns
	#
	if ($maildom ne $fqdn) {
		if (defined($MailDomByBasename{$maildom})) {
			return "maildom $maildom is already a basename";
		}
		if ($maildom !~ /\.$fqdn$/) {
			return "WARNING: maildom $maildom isn't in $fqdn";
		}
		$MailDomByBasename{join($;, $`, $fqdn)} = $maildom;
#D		print STDERR "(maildombybasename{$`, $fqdn} <- $maildom)\n";
	}
	return '';
} #sub


########################################################################
##
## complain
##
## print barf messages.
##
sub complain {
	local($message, $fatality) = @_;

	return unless ($fatality || $WorkFile || $opt_f);
	printf STDERR "$filename:$.:\n    $message (%s)\n\t$_\n",
		$fatality ?"fatal error" :"ignored";
	$ExitStatus |= $fatality;
}


########################################################################
##
## dump_it_out
##
## Print out the entire database for debugging.
##

sub dump_it_out {
	print "--- FwdZones:\n";
	foreach $zone (sort(keys(%FwdZones))) {
		printf("[%s] ", &pn($zone));
	}
	print "\n";
	print "--- RevZones:\n";
	foreach $zone (sort(keys(%RevZones))) {
		printf("[%s] ", &sends'fmtAddr($zone));
	}
	print "\n";
	print "--- BasenameByAddr:\n";
	while (($key,$val) = each(%BasenameByAddr)) {
		printf("[%s] -> '%s'\n", &sends'fmtAddr($key), &pn($val));
	}
	print "--- BasenameByAlias:\n";
	while (($key,$val) = each(%BasenameByAlias)) {
		printf("'%s' -> '%s'\n", &pn($key), &pn($val));
	}
	print "--- AddrsByAlias:\n";
	while (($key,$val) = each(%AddrsByAlias)) {
		printf("'%s' ->", &pn($key));
		foreach (split(/$;/, $val)) {
			printf(" [%s]", &sends'fmtAddr($_));
		}
		printf("\n");
	}
	print "--- AddrsByBasename:\n";
	while (($key,$val) = each(%AddrsByBasename)) {
		printf("'%s' ->", &pn($key));
		foreach (split(/$;/, $val)) {
			printf(" [%s]", &sends'fmtAddr($_));
		}
		print "\n";
	}
	print "--- AliasesByBasename:\n";
	while (($key,$val) = each(%AliasesByBasename)) {
		printf("'%s' -> (", &pn($key));
		foreach (split(/\s/, $val)) {
			printf(" %s", &pn($_));
		}
		print " )\n";
	}
	print "--- AliasesByAddr:\n";
	while (($key,$val) = each(%AliasesByAddr)) {
		printf("[%s] -> (", &sends'fmtAddr($key));
		foreach (split(/\s/, $val)) {
			printf(" %s", &pn($_));
		}
		print " )\n";
	}
	print "--- SortedAddrs:\n";
	$x = 0;
	foreach (@SortedAddrs) {
		printf(" [%s]", &sends'fmtAddr($_));
		if (!(++$x % 4)) {
			print "\n";
			$x = 0;
		}
	}
	print "\n" if $x;
}


########################################################################
##
## pn
##
## Convert all the perl-separator characaters ($;) in a string to the 
## given seperator (or ':' if non-specified).
##

sub pn {
	local($name, $sep) = @_;

	$sep = ':' unless defined $sep;
	return join($sep, split($;, $name));
}


########################################################################
##
## produce_hostdb
##
## Produce the two host databases. The hostoptions file, which is a
## hostname to options mapping, and the host domain file, which is a
## hostname to mail-domain mapping

sub produce_hostdb {
	local($basename, $options, $maildom);
	local($_, *opt, *dom, %generated);

	#
	# First generate the hostoptions file. Format is:
	#	host.fqdn	hostoptions
	# e.g.
	#	venus.sjc.hw.sony.com   -d-
	# the output includes basenames and aliases
	#
	%generated = ();
	open(opt, ">$HostOpt.$$") || die "produce_hostdb: $HostOpt.$$: $!";
	print opt "# $DoNotEdit\n";
	while (($basename, $options) = each(%OptionsByBasename)) {
		printf opt "%s\t-%s-\n", &pn($basename, '.'), $options;
		$generated{$basename} = '';
		foreach (split(/\s/, $AliasesByBasename{$basename})) {
			next if defined($generated{$_});
			printf opt "%s\t-%s-\n", &pn($_, '.'), $options;
			$generated{$_} = '';
		}
	}
	close(opt) || die "produce_hostdb: close($HostOpt.$$): $!";
	&sends'renameOrUnlink("$HostOpt.$$", "$HostOpt");

	#
	# Now do the hostdomain file. Format is:
	#	basename.fqdn	maildomain
	# e.g.
	#	cons1.sjc.in.sel.sony.com       in.sel.sony.com
	# basenames only.
	open(dom, ">$HostDom.$$") || die "produce_hostdb: $HostDom.$$: $!";
	print dom "# $DoNotEdit\n";
	while (($basename, $maildom) = each(%MailDomByBasename)) {
		printf dom "%s\t%s\n", &pn($basename, '.'), $maildom;
		# Handle ugly MSMail hosts here.
		if ( $OptionsByBasename{$basename} =~ /p/ ) {
			foreach $a (split(' ',$AliasesByBasename{$basename})) {
				printf dom "%s\t%s\n",&pn($a,'.'),$maildom ;
			}
		}
	}
	close(dom) || die "produce_hostdb: close($HostDom.$$): $!";
	&sends'renameOrUnlink("$HostDom.$$", "$HostDom");
}


########################################################################
##
## produce_fwddns 
##
## Generate the FWD zone DNS file for the FQDN. Note a FQDN can contain
## more than one mail domain.
##

sub produce_fwddns {
	local($fqdn) = @_;
	local($filename) = $sends'zoneDir .'/work/'. $fqdn;
	local($_, $f, $relay, $mxrr, $name, $nameFqdn, $addr, @addrs,
		$basename, $alias, $maildom, $firstMX);

	print STDERR "Producing fwd_dns for $fqdn\n" if $opt_d ;
	open(f, ">$filename") || die "produce_fwddns: $filename: $!";

	%MxdThisName = ();
	print f "; $DoNotEdit\n";
	printf f "\$ORIGIN %s.\n", $fqdn;

	# 
	# For each of the maildomains within the FQDN zone, we set up MX's
	# pointing at the mail relays for the zone.
	#
	foreach $maildom (split(/$;/, $MailDomsByZone{$fqdn})) {
		if ($maildom !~ /$fqdn$/) {
			print STDERR
				"$0: PANIC: $maildom does not end in $fqdn\n";
			exit(1);
		}
		$_ = $`;
		if (/\.$/) { # mail domain is within fqdn
#D			printf STDERR
#D				"(calling genMX for $`, fqdn $fqdn)\n";
			&genMX('f', $`, $`, 1, $fqdn);
		} else {  # maildomain == fqdn
			&genMX('f', '@', '@', 1, $fqdn);
		}
	}

	#
	# Now do the A records for local host and the A & MX records for
	# all the iaddrs. Also PTR records for networks.
	# 
	printf f "%-31s\tIN A\t%s\n",
		'localhost',
		'127.0.0.1';
 addr:	foreach $addr (@SortedAddrs) {
		$basename = $BasenameByAddr{$addr};
		($name, $nameFqdn) = split($;, $basename);
		next unless ($nameFqdn eq $fqdn);

		# if this address is an RFC1101 network name then generate
		# the proper PTR record.
		if ( defined ($is1101{$addr}) ) {
			# host .0 is an RFC1101 network name
			printf f "%-31s\tIN PTR\t%s\n",
				$name, &sends'revAddr($addr);
			next addr;
		}
		printf f "%-31s\tIN A\t%s\n", $name, &sends'fmtAddr($addr);

		# 
		# The Ugle Ugly Ugly MSMail kludge. We generate A records
		# for each of the aliases the MSMail postorifice has.
		#
		if ($OptionsByBasename{$basename} =~ /p/) {
			# MSMail mess. Need to generate A records for each 
			# alias.
			foreach $a (split(' ',$AliasesByBasename{$basename})) {
				local ($aname) = split($;,$a) ;
				printf f "%-31s\tIN A\t%s\n", 
					$aname, &sends'fmtAddr($addr);
				
			}
		}

		#
		# no MX RR's for this host if it's dumb, since we will not
		# be qualifying the address with @domain in SMTP and it would
		# be Very Bad to send such a thing to any host but "this" one.
		#
		if ($OptionsByBasename{$basename} !~ /d/) {
			#
			# figure out whether the host has the "-g" option
			# since if it does, we need to tell &genMX() not to
			# generate a cost=0 mx "to self".
			#
			$firstMX = ($OptionsByBasename{$basename} =~ /g/)
					? 1
					: 0;
			&genMX('f', $name, '', $firstMX, $fqdn);
		}

		#
		# If this iaddr has more than one name (is a multi-homed
		# host), then we handle aliases differently
		#
		@addrs = split(/$;/, $AddrsByBasename{$basename});
		if ($#addrs > 0) {
			#
			# multi-homed host -- aliases are A RR's
			#
			foreach $alias (split(/\s/, $AliasesByAddr{$addr})) {
				($_) = split($;, $alias);
				printf f "%-31s\tIN A\t%s\n",
					$_, &sends'fmtAddr($addr);
				&genMX('f', $_, '', 0, $fqdn);
			}
		} else {
			#
			# single-homed host -- aliases are CNAME RR's
			# except for MSMail hosts
			#
			if ($OptionsByBasename{$basename} !~ /p/) {
				foreach $alias (split(/\s/,
						      $AliasesByBasename{$basename})) {
					($_) = split($;, $alias);
					printf f "%-31s\tIN CNAME %s\n",
						$_, $name;
				}
			}
		}
	}

	#
	# Now generate the MX records specified by the special comment
	# syntax
	#
	foreach $key (sort(keys(%mxList))) {
		local ($base,$tail) = split($;,$key) ;
		local ($first) = 1 ;

		next unless ( $tail eq $fqdn ) ;

		local (@targets) = split ($;,$mxList{$key}) ;
		while (@targets) {
			local ($pri) = shift (@targets) ;
			local ($target) = shift (@targets) ;

			printf f "%-31s\tIN MX\t%d %s\n",
					($first ? $base : "") ,$pri,$target ;
			$first = 0 ;
		}
	}

	close(f) || die "produce_fwddns: close($filename): $!";
}

###############################################################################
##
## genMX
##
## Generate the MX records for a name.

sub genMX {
	local($fh, $name, $printname, $firstCost, $fqdn) = @_;
	local($oldfh, $relay, $mxrr, $format, $f, $maildom);

	return if defined($MxdThisName{join($;, $name, $fqdn)});
	$MxdThisName{join($;, $name, $fqdn)} = '';

	$oldfh = select($fh);

	if ($printname) {
		$format = '%-31s';
	} else {
		$format = '%s';
		$printname = "\t\t\t";
	}

	#
	# name == '@' if maildomain == fqdn (in caller)
	#
	if ($name eq '@') {
		$maildom = $fqdn;
	} else {
		#
		# Get the maildomain based on the host name or (if we're
		# processing an alias), the basename for the alias.
		#
		$maildom = $MailDomByBasename{join($;, $name, $fqdn)};
		if (!$maildom) {
			$maildom = $MailDomByBasename{
					$BasenameByAlias{
						join($;, $name, $fqdn)
					}
				};
		}
	}
	if (!$maildom) {
		print STDERR "$0: PANIC: ",
			"!maildom; fqdn is \"$fqdn\"; name is \"$name\"\n";
		exit(1);
	}

	#
	# Generate an MX pointing to the host itself if needed
	#
	if ($firstCost == 0) {
		printf "$format\tIN MX\t%d %s\n", $printname, 0, $name;
		$format = '%s';
		$printname = "\t\t\t";
	}

	#
	# Now generate MX's pointing at all the mailrelays for this
	# maildomain
	#
	$f = 1;
	foreach $relay (split($;, $RelaysByMailDom{$maildom})) {
		foreach $mxrr (split(/\s+/, $relay)) {
			printf "$format\tIN MX\t%d %s\n",
				$printname, $f, $mxrr;
			$format = '%s';
			$printname = "\t\t\t";
		}
		$f++;
	}

	select($oldfh);
}


########################################################################
##
## produce_revdns
##
## Generate the reverse-lookup data file for the zone.

sub produce_revdns {
	local($inet) = @_;
	local($ZoneSize) = $RevZones{$inet} ;

## DEBUG
	print "Zone size for ", &sends'fmtAddr ($inet), " is ",
		$ZoneSize, "\n" if $opt_d ;

	local(@net) = (split(/\./, &sends'fmtAddr($inet)))[0..($ZoneSize-1)];
	local($net) = join('.', @net);
	local($filename) = $sends'zoneDir .'/work/'. $net;
	local(@addrs);
	local($iaddr, $_, $basename);

	print STDERR "Producing rev_dns for $net ($filename)\n" if $opt_d ;
	open(f, ">$filename") || die "produce_revdns: $filename: $!";

	print f "; $DoNotEdit\n";
	printf f "\$ORIGIN %s\n", &sends'revAddr($inet, $ZoneSize);
	foreach $iaddr (@SortedAddrs) {

## something's wrong with SortedAddrs--the keys it was generated from seem
## to be sorted as if they were ascii.
##
#		print "Looking at address: ",
#			&sends'fmtAddr ($iaddr), " (", $iaddr, ")\n" 
#			if $opt_d ;
#		# optimization only
#		last if (($iaddr & $ZoneMaskByAddr{$iaddr}) > $inet) ;

		next unless (($iaddr & $ZoneMaskByAddr{$iaddr}) == $inet) ;
		$basename = $BasenameByAddr{$iaddr};
		printf f "%-31s\tIN PTR\t%s\n",
			&sends'revAddr($iaddr), &pn($basename, '.').'.';

		#
		# Handle a network name (i.e. host == *.0)
		#
		if ( defined ($is1101{$iaddr}) ) {
			# host .0 is an RFC1101 network name
			printf f "\t\t\t\tIN A\t%s\n",
				&sends'fmtAddr($is1101{$iaddr}); # XXX iaddr?
		}

		#
		# Handle multi-homed hosts
		#
		@addrs = split(/$;/, $AddrsByBasename{$basename});
		if ($#addrs > 0) {
			# multi-homed host -- ptr's to match a rr's
			foreach (split(/\s/, $AliasesByAddr{$iaddr})) {
				printf f "%-31s\tIN PTR\t%s\n",
					'', &pn($_, '.').'.';
			}
		}
	}
	close(f) || die "produce_revdns: close($filename): $!";
}



######################################################################
##
## Section related to mapping a IP address to a Netmask, Zonemask or
## Zonesize paramater.
##

@NetWorks = () ;
@NetRangeMasks = () ;
@ZoneSizes = () ;
@ZoneMasks = () ;
@RFCMasks = () ;
$NetWorksOK = 0 ;   # set to 1 after above arrays are set up.

#
# cache the Config values for the networks.
#
sub checkGlobals {
	if ( ! $NetWorksOK ) {
		local (@cInfo) = split (' ',$sends'Config{'network'}) ;
		local ($f,$iaddr,$zonesize,$zonemask) ;
		

		while ( $#cInfo - $[ + 1 > 0 ) {
			($f,$iaddr) = &sends'iaddrOf(shift (@cInfo)) ;
			die "$f\n" if $f ;
			push (@NetWorks,$iaddr) ;

			($f,$iaddr) = &sends'iaddrOf(shift (@cInfo)) ;
			die "$f\n" if $f ;
			push (@NetRangeMasks,$iaddr) ;

			($f,$iaddr) = &sends'iaddrOf(shift (@cInfo)) ;
			die "$f\n" if $f ;
			push (@RFCMasks,$iaddr) ;

			$zonesize = shift (@cInfo) ;
			push (@ZoneSizes,$zonesize) ;

			$zonemask = 0 ;
			$zonemask = 0xff000000 if ($zonesize == 1) ;
			$zonemask = 0xffff0000 if ($zonesize == 2) ;
			$zonemask = 0xffffff00 if ($zonesize == 3) ;
			die "Bogus zonesize in config: $zonesize\n" 
				if ! $zonemask ;

			push (@ZoneMasks,$zonemask) ;
		}

		$NetWorksOK = 1 ;
	}
}


sub inetToZoneSize {
	local ($inet) = @_ ;
	
	&checkGlobals ;

	for ($i = $[ ; $i <= $#NetWorks ; $i++) {
		if (($NetWorks[$i] & $NetRangeMasks[$i]) == 
		    ($inet & $NetRangeMasks[$i]))   {
			return ("", $ZoneSizes [$i]) ;
		}
	}

	return ("Didn't find a zonesize for " . &sends'fmtAddr ($inet), 0) ;
}


#
# Generate the global level zone mask based on the network zonesize config 
# entry.
#
sub inetToZoneMask {
	local ($inet) = @_ ;
	
	&checkGlobals ;

	for ($i = $[ ; $i <= $#NetWorks ; $i++) {
		if (($NetWorks[$i] & $NetRangeMasks[$i]) == 
		    ($inet & $NetRangeMasks [$i])){
			return ("", $ZoneMasks [$i]) ;
		}
	}

	return ("Didn't find a zonemask for " . &sends'fmtAddr ($inet), 0)
}


########################################################################
##
## Map an inet address into a netmask based ONLY on the config file values
## (i.e. NOT the control files as well).
## Expects the argument to be an integer, not a dotted quad.

#
# CURRENTLY NOT USED
#

# sub inetToGlobalNetMask {
# 	local ($inet) = @_ ;
# 	
# 	&checkGlobals ;
# 
# 	for ($i = $[ ; $i <= $#NetWorks ; $i++) {
# 		if (($NetWorks[$i] & $NetRangeMasks[$i]) == 
#		    ($inet & $NetRangeMasks[$i])){
# 			return ('', $NetRangeMasks[$i]) ;
# 		}
# 	}
# 
# 	return ("Didn't find a netmask for " . &sends'fmtAddr ($inet), 0) ;
# }


#
# Get the RFC1101 mask from the global config file for this address.
#
sub inetToRFC1101 {
	local($inet) = @_ ;

	&checkGlobals ;

	for ($i = $[ ; $i <= $#NetWorks ; $i++) {
		if (($NetWorks [$i] & $NetRangeMasks [$i]) == ($inet & $NetRangeMasks [$i])){
			return ('', $RFCMasks [$i]) ;
		}
	}

	return ("Didn't find a RFC1101 mask for " . &sends'fmtAddr ($inet), 0) ;
}

	
#
# CURRENTLY NOT USED
#

# sub naturalMask {
# 	local ($inet) = @_ ;
# 	local ($classA) = (($inet & 0x80000000) == 0);
# 	local ($classB) = (($inet & 0xc0000000) == 0x80000000);
# 	local ($classC) = (($inet & 0xe0000000) == 0xc0000000);
# 
# 	return 0xff000000 if $classA ;
# 	return 0xffff0000 if $classB ;
# 	return 0xffffff00 if $classC ;
# 
# 	die "No natural mask?\n" ;
# }

