#!/usr/bin/perl

# Copyright (c) 1994 by Vixie Enterprises
# 
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies, and that
# the name of Vixie Enterprises not be used in advertising or publicity
# pertaining to distribution of the document or software without specific,
# written prior permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND Vixie Enterprises DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL Vixie Enterprises
# BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
# OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

# $Id: do-users.pl,v 1.10 1995/02/16 12:45:43 brister Exp $

# Credits:
#       Paul Vixie <paul@vix.com> (functional design partner; original coder)
#       Jerry Scharf <scharf@sony.com> (functional design partner)
#       James Brister <brister@vix.com> (major code rework; maintainance)



#
# Generates the user database. There is a separate database per MAILDOMAIN.
# So if multiple workgroups have the same maildomain then their entries are
# merged.

#
# Boilerplate
#

$sends'Root = $ENV{'SENDS_ROOT'} ;
$sends'Root = "/usr/local/sends" unless defined ($sends'Root) ;
require "$sends'Root/lib/sends-config.pl" ;

##################################################

$MySep     = "\016";
$DoNotEdit = "# this file was automatically generated - do not edit\n";

# Keys that are not allowed to be in the default entry in the user file.
@NoDefault = split(' ',$sends'Config{'user_key_no_defaults'});

# Can't have more than one field in an entry with any of these keys.
@NoDups    = split(' ',$sends'Config{'user_key_no_dups'}) ;

# Got to have at least these fields specified in an entry.
@MinFields = split(' ',$sends'Config{'user_key_min_fields'}) ;

$MissingOk = 1 ;   # flag to some calls of get_input
$ExitStatus = 0;

#
# get subdomain name and list of no-no domains
#
&parse_options();

$sends'Debug = $sends'Config{'debug'} ;
$sends'Debug = $opt_d unless $sends'Debug ;

##
## if the -f flag is used then we iterate over all domains. Otherwise we 
## only process a particular group's file.
##

if ($opt_f) {
	%ProducedOutput = ();
##	$,=':'; print STDERR "domains: (", (&sends'DomainList()), ")\n";

	foreach $WorkDomain (&sends'DomainList()) {

##	    $,=':'; print STDERR "groups($WorkDomain): (",
##			  (&sends'GroupList($WorkDomain)),
##			  ")\n";

	    foreach $WorkGroup (&sends'GroupList($WorkDomain)) {
		$MailDom = &sends'MailDomByGroup($WorkDomain, $WorkGroup);
		print STDERR "[$WorkDomain,$WorkGroup]'s maildom is $MailDom\n"
			if $opt_d;
		if (!defined($MailDom)) {
			die "MailDomByGroup FAILED: ".$sends'Error."\n";
		}
		next if defined $ProducedOutput{$MailDom};
		print STDERR "[force: $MailDom]\n"		if $opt_d;
		%KeysUsed = ();
		@Users = ();

		$inputFile = &sends'GroupToPath($WorkDomain, $WorkGroup)
				. '/output/users' ;

		# a missing inputFile could mean this is a new group that
		# has never had a file processed
		next if ( ! -f $inputFile ) ;

		# 
		# First we find all the other Groups that have the same
		# MAILDOMAIN as the group we're processing now. And get
		# their data.
		#
		&sends'forEachGroup("&main'doNonworkGroup()");

		#
		# Now get the data for our group itself.
		#
		$_ = &get_input($inputFile);
		if ($_) {
			warn "$_\n";
		} else {
			&produce_userdb($MailDom);
			$ProducedOutput{$MailDom} = '';
		}
	    }
	}
} else {
	# 
	# We're only processing one workgroup in this run
	#
	$MailDom = &sends'MailDomByGroup($WorkDomain, $WorkGroup);
	if (!defined($MailDom)) {
		print STDERR "MailDomByGroup FAILED: ".$sends'Error."\n";
		exit(1);
	}

	%KeysUsed = ();
	@Users = ();
	
	#
	# Find and process all the other groups that share the maildomain 
	# with the group we're processing.
	#
	&sends'forEachGroup("&main'doNonworkGroup()");

	print STDERR "(new: $MailDom)\n"                   	if $opt_d;

	#
	# Now get the input for our specific group.
	#
	$_ = &get_input(&sends'GroupToPath($WorkDomain,$WorkGroup)
			. '/work/users');
	die "$_\n" if $_;

	&dump_it_out()						if $opt_d;

	#
	# $ExitStatus is set by &complain, and the only zero-valued
	# complaint is a MAILHOST that doesn't have an A record. All others
	# are fatal.
	#

	if ((!$opt_n) && (!$ExitStatus)) {
		&produce_userdb($MailDom);
	}
}

exit($ExitStatus);

############################################################################
##
## doNonWorkGroup
##
## Called via an eval inside of sends'forEachGroup in which $sends'Domain
## and $sends'Group will be set.
##
## We find all the groups and domain combinations that have the same
## maildomain as the one we're processing at the top level ($MailDom) and
## REMIND 
##
## 
sub doNonworkGroup {
	local($maildom) = &sends'MailDomByGroup($sends'Domain, $sends'Group);
	local($_);

	if (!defined($maildom)) {
		print STDERR "MailDomByGroup FAILED: ".$sends'Error."\n";
		return 0;
	}
	if ($opt_d) {
		print STDERR
		"doNonWork:\n" .
			"\tsends [$sends'Domain, $sends'Group, $maildom]\n" .
			"\twork: [$WorkDomain, $WorkGroup, $MailDom]\n";
	}
	return 1 if ($sends'Domain eq $WorkDomain)
		 && ($sends'Group eq $WorkGroup);
	return 1 if $maildom ne $MailDom;
	$_ = &get_input($sends'Path.'/output/users',$MissingOk);
	warn "$_\n" if $_;
	return 1;
}

############################################################################
##
## parse_options
##
## The usual.
##

sub parse_options {
	$opt_w = $opt_W = $opt_f = $opt_n = $opt_d = undef;
	&Getopts('w:W:fnd');
	if ((!$opt_w != !$opt_W)
	 || (!$opt_f == !$opt_w)) {
		print STDERR
			"usage: $0 { -W wrkdom -w wrkgrp | -f } [-n] [-d]\n";
		exit(1);
	}
	($WorkDomain, $WorkGroup) = ($opt_W, $opt_w);
}

############################################################################
##
## get_input
##
## Read in the specified FILENAME. The file is in MailHeader format. An
## entry with the key 'default' is saved and used as the default values for
## missing fields from follwing entries.
##

sub get_input {
	local($filename,$missingOk) = @_;
	local(%default, %entry, @rr);
	local($_, $key, $val, $u, $d);
	local(*f);

	%default = ();
	print STDERR "get_input($filename)\n"			if $opt_d;
	if ( ! open(f, $filename) ) {
		if ( $missingOk ) {
			return '' ;
		} else {
			return "$filename: $!";
		}
	}

  user:	for (;;) {
		%entry = &sends'parseHdrs("main'f");
		if (!scalar(%entry)) {
			&complain($sends'Error, 1) if defined($sends'Error);
			last user;
		}

		#
		# process default entry if this is one
		#
		if (defined($entry{'default'})) {
			$_ = $entry{'default'};
			delete $entry{'default'};
			if ($_ eq 'replace') {
				%default = %entry;
			} elsif ($_ eq 'merge') {
				while (($key, $val) = each(%entry)) {
					$default{$key} = $val;
				}
			} else {
				&complain("bad \"default\" field: $_", 1);
				last user;
			}
			foreach $key (@NoDefault) {
				if (defined($default{$key})) {
					&complain(
					    "can't specify def'lt for: $key",
					    1);
					last user;
				}
			}
			next user;
		}

		#
		# apply defaults to this entry
		#
		while (($key, $val) = each(%default)) {
			$entry{$key} = $val unless defined $entry{$key};
		}

		#
		# make sure required fields are all there
		#
		foreach $key (@MinFields) {
			if (!defined($entry{$key})) {
				&complain("missing required field: $key", 1);
				last user;
			}
		}

		#
		# make sure that no must-be-unique keys are duplicated
		#
		foreach $key (@NoDups) {
			if (defined($KeysUsed{$key,$entry{$key}})) {
				&complain("duplicate $key: $entry{$key}", 1);
				last user;
			}
			$KeysUsed{$key,$entry{$key}} = '';
		}

		#
		# see if mailhome is reasonable
		# REMIND. Should cache the digs maybe?
		#
		@rr = &sends'dig($entry{'mailhome'}, 'a');
		print STDERR "\tA: $#rr\n"	if $opt_d;
		if ($#rr < $[) {
			&complain("mailhome unmailable", 0);
		}

		#
		# store this entry in memory for later database dump
		#
		push(@Users, join($MySep, %entry));
	} #user
	close(f);
	return '';
} #sub


########################################################################
##
## complain
##
## Spit out an error message and set a global variable describing to
## seriousness of the error.
##

sub complain {
	local($message, $fatality) = @_;
	local($key, $val);

#	return unless ($fatality || $opt_f);
	printf STDERR "$filename:$.:\n    $message (%s)\n",
		$fatality ?"fatal error" :"ignored";
	while (($key,$val) = each(%entry)) {
		printf STDERR "\t%s:\t%s\n", $key, $val;
	}
	$ExitStatus |= $fatality;
}

########################################################################
##
## dump_it_out
##
## Take the Users array generated earlier and spit out the 
## MailHeader-style database of users to STDERR.
##

sub dump_it_out {
	local($user, %key, $val);
	local(%entry);

	print STDERR "--- Users:\n";
	foreach $user (@Users) {
		%entry = split(/$MySep/, $user);
		while (($key,$val) = each(%entry)) {
			printf STDERR "\t'%s' -> %s\n", $key, $val;
		}
		print STDERR "\n";
	}
}

########################################################################
##
## produce_userdb
##
## spit out the user database for the MAILDOMAIN from the input processed
## earlier. 
##

sub produce_userdb {
	local($maildom) = @_;
	local($user, %key, $val);
	local(%entry);
	local(*f);

	$f = "$sends'userDir/work/$maildom";
	print STDERR "produce_userdb($maildom): $f\n"		if $opt_d;
	open(f, ">$f") || die "$f: $!";
	print f $DoNotEdit;
	foreach $user (@Users) {
		%entry = split(/$MySep/, $user);
		while (($key,$val) = each(%entry)) {
			printf f "%s: %s\n", $key, $val;
		}
		print f "\n";
	}
	close(f) || die "close($f): $!";
}
