#!/usr/bin/perl

# Copyright (c) 1994 by Vixie Enterprises
# 
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies, and that
# the name of Vixie Enterprises not be used in advertising or publicity
# pertaining to distribution of the document or software without specific,
# written prior permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND Vixie Enterprises DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL Vixie Enterprises
# BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
# OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

# $Id: mk-users.pl,v 1.7 1995/02/16 12:45:43 brister Exp $

# Credits:
#       Paul Vixie <paul@vix.com> (functional design partner; original coder)
#       Jerry Scharf <scharf@sony.com> (functional design partner)
#       James Brister <brister@vix.com> (major code rework; maintainance)

#
# Scan all the work directories for files to process. The files would have 
# been put here by 'scan-ftp'.
#

# Boilerplate

$sends'Root = $ENV{'SENDS_ROOT'} ;
$sends'Root = "/usr/local/sends" unless defined ($sends'Root) ;
require "$sends'Root/lib/sends-config.pl" ;

&getLock ;

#
# Here we go....
#

$Debugging = $sends'Config{'debug'} ;
$Debugging = 0 unless defined $Debugging ;

$DoNotEdit = "# this file was automatically generated - do not edit\n";

&parse_options();

$opt_d = 1 if !defined ($opt_d) && $Debugging ;

chdir($sends'userDir) || die "chdir($sends'userDir): $!";

# If the '-f' flag is used then the standard work directory is not 
# processed, but the previously processed files (done on a previous run 
# of this file) are re-processed.
#
# This option is currently disabled

if ($opt_f) {
	$workDir = 'output';
} else {
	$workDir = 'work';
}

%MailDomList = ();
foreach (&sends'MailDomList()) {
	$MailDomList{$_} = '';
}

opendir(workDir, $workDir) || die "$sends'userDir/$workDir: $!";
$somethingHappened = 0;
while ($workFile = readdir(workDir)) {
	next if $workFile =~ /^\./;
	next unless defined $MailDomList{$workFile};
	&processWorkFile($workFile);
	$somethingHappened++;
}
closedir(workDir);
exit 0 unless $somethingHappened || $opt_f;

&produceUserFiles() unless $opt_n;

exit(0);

############################################################################
##
## parse_options
##
## Parse the legal options for this program... -n -f -d.
##
##	-d == debug
##	-f == force re-processing
##	-n == don't process file, just get them ready (i.e. move them into
##	      the processing directory.

sub parse_options {
	$opt_n = $opt_d = $opt_f = undef;
	&Getopts('ndf');
	die "the -f option has been disabled\n"	if $opt_f;
}


############################################################################
##
## processWorkFile
##
## Look at the file in the subdirectory 'work' and if it is not the same as
## the version that exists in the subdirectory 'output' then move it to the
## output directory.
##
## The -f flag forces the files in the the output directory to be
## considered new.

sub processWorkFile {
	local($file) = @_;

	if (!$opt_f) {
		if (0 == (system("cmp -s work/$file output/$file") >> 8)) {
			# file has not changed
			unlink "work/$file";
			return;
		}
	}
	# file has changed, or one of them doesn't exist, or whatever.

	print STDERR "processWorkFile($file): has changed\n"	if $opt_d;

	# do this last for atomicity and restartabililty
	return if $opt_f;
	rename("work/$file", "output/$file") || die "rename(work/$file): $!";
}


############################################################################
##
## produceUserFiles
##
## Do the actual grinding of the data files....
## Iterates over all the MAIL-DOMAINS of all the groups and generates the
## User files. A master file containing the full user information for all
## the groups combined is produced from the user files in the group
## directories.
##
## A master sendmail alias file (sort of) is also produced. This is used on
## the gateway machines (regional routers actually) to the username -> Full
## name mapping on mail headers.
##
sub produceUserFiles {
	local(*userfile, *mailaddr, *fulltext);
	local(%entry);
	local($_, $key, $val, $mailhomeuser);
	local(@mafl) = ('username', 'mailname', 'mailhome');

	open(fulltext, ">fulltext.$$") || die "fulltext.$$: $!";
	open(mailaddr, ">mailaddr.$$") || die "mailaddr.$$: $!";

	print mailaddr $DoNotEdit;
	print fulltext $DoNotEdit;

	foreach $fqdn (keys(%MailDomList)) {
		next unless open(userfile, "<output/$fqdn");
		print STDERR "producing for '$fqdn'\n"		if $opt_d;
  entry:	for (;;) {
			%entry = &sends'parseHdrs("main'userfile");
##			print STDERR "#/entry = ", scalar(%entry), "  $!\n";
			last unless scalar(%entry);

			while (($key,$val) = each(%entry)) {
				print fulltext "$key: $val\n";
			}
			print fulltext "maildom: $fqdn\n";
			print fulltext "\n";

			# 
			# Make sure we have our minimum required key set.
			#
			foreach $key (@mafl) {
				if (!defined($entry{$key})) {
					local($,) = '/';

					print STDERR
						"missing <$key> field in ",
						(keys(%entry)),
						"\n";
					next entry;
				}
			}

			# 
			# Generate the sendmail database entry.
			#
			$mailhomeuser = ($entry{'username'} eq '*')
					? '%s'
					: $entry{'username'};
			printf mailaddr "%s<@%s> %s<@%s> %s<@%s>\n",
				$entry{'username'}, $fqdn,
				$entry{'mailname'}, $fqdn,
				$mailhomeuser, $entry{'mailhome'};
		} #user
		close(userfile);
	} #fqdn

	close(fulltext) || die "close(fulltext.$$): $!";
	close(mailaddr) || die "close(mailaddr.$$): $!";

	rename("fulltext.$$", "fulltext") || die "rename(fulltext.$$): $!";
	rename("mailaddr.$$", "mailaddr") || die "rename(mailaddr.$$): $!";
}
