#!/usr/bin/perl

# Copyright (c) 1994 by Vixie Enterprises
# 
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies, and that
# the name of Vixie Enterprises not be used in advertising or publicity
# pertaining to distribution of the document or software without specific,
# written prior permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND Vixie Enterprises DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL Vixie Enterprises
# BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
# OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

# $Id: mk-zones.pl,v 1.12 1995/05/25 14:02:52 brister Exp $

# Credits:
#       Paul Vixie <paul@vix.com> (functional design partner; original coder)
#       Jerry Scharf <scharf@sony.com> (functional design partner)
#       James Brister <brister@vix.com> (major code rework; maintainance)

#
# Generate the Zone files.
# 

#
# Boilerplate
#

$sends'Root = $ENV{'SENDS_ROOT'} ;
$sends'Root = "/usr/local/sends" unless defined ($sends'Root) ;
require "$sends'Root/lib/sends-config.pl" ;

&getLock ;

##################################################

$DoNotEdit = "; this file was automatically generated - do not edit\n";

%SOA = (
	'masterhost',	$sends'Config{'zone_master'},
	'hostmaster',	$sends'Config{'hostmaster'},
	'refresh',	$sends'Config{'zone_refresh'},
	'retry',	$sends'Config{'zone_retry'},
	'expire',	$sends'Config{'zone_expire'},
	'minimum',	$sends'Config{'zone_minimum'},
	'', ''
);

$kickNamed = 0 ;  # we kick named if files change.

$PriAddr = $sends'Config{'primary'} ;

@NameServers = ( split (' ',$sends'Config{'name_servers'}) ) ;

# -f means force a regeneration, -d is debug and -n means don't 
# actually grind any files.

&parse_options();

$opt_d = 1 if $sends'Config{'debug'} ;


%FileByZone = ();
chdir($sends'zoneDir) || die "chdir($sends'zoneDir): $!";

if ($opt_f) {
	$workDir = 'output';
} else {
	$workDir = 'work';
}

#
# look in the work directory for stuff to do..
#
opendir(workDir, $workDir) || die "$sends'zoneDir/$workDir: $!";
$somethingHappened = 0;
while ($workFile = readdir(workDir)) {
	next if $workFile =~ /^\./;
	next if $opt_f && ($workFile =~ /\.data$/);
	&processWorkFile($workFile);
	$somethingHappened++;
}
closedir(workDir);
exit unless $somethingHappened || $opt_f;

&produceBootFiles();

&kickDNS if ( $kickNamed ) ;

exit(0);

############################################################################
##
## Send the named server a SIGHUP 
##

sub kickDNS {
	local ($pidfile) = $sends'Config{'named-pid-file'} ;
	$pidfile = "/etc/named.pid" unless $pidfile ;

	local ($pid,$kickRes) ;

	if ( -e $pidfile ) {
		open (PID,"<$pidfile") || die "open ($pidfile): $!" ;
		chop ($pid = <PID>) ;
		close (PID) ;
		$kickRes = `kill -HUP $pid` if $pid ;
#		print STDERR "reloading nameserver\n" ;
	} else {
		print STDERR "No NAMED pid file ($pidfile). Can't kick.\n" ;
	}

	1 ;
}
############################################################################
##
## standard option parsing.
##

sub parse_options {
	$opt_n = $opt_d = $opt_f = undef;
	&Getopts('ndf');
	die "the -f option has been disabled\n" if $opt_f;
}


############################################################################
##
##  processWorkFile
##
## For each zone data file in the work directory generate a DNS SOA file that
## will $INCLUDE the datafile. Move the SOA file and the data file into the
## output directory.
##

sub processWorkFile {
	local($file) = @_;
	local($zone);

	if ($file =~ /^[\d\.]+$/) {
		$zone = join('.', reverse(split(/\./, $file)), 'in-addr.arpa');
	} else {
		$zone = $file;
	}
	print STDERR "processWorkFile($file)... zone: $zone\n"	if $opt_d;
	$FileByZone{$zone} = $file;
	if (!$opt_f) {
		if (0==(system("cmp -s work/$file output/$file.data") >> 8)) {
			print STDERR "\tfile has not changed\n"	if $opt_d;
			unlink "work/$file";
			return;
		}
	}
	print STDERR "\tfile has changed or does not exist\n"	if $opt_d;

	$kickNamed = 1 ;

	&makeSOAfile($file, $zone);

	# do this last for atomicity and restartabililty
	return if $opt_f;
	rename("work/$file", "output/$file.data") || die "rename(work/$file): $!";
}

############################################################################
##
## makeSOAfile
##
## Generate the SOA record and the proper ORIGIN and INCLUDE statements for
## bind to handle this zone's data.
##

sub makeSOAfile {
	local($file, $zone) = @_;
	local(*soa);
	local($_, $version, $opened);
	local($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =
		localtime(time);
	local($nowstr) = sprintf("%4d%02d%02d", $year+1900, $mon+1, $mday);

	#
	# assume that the serial number won't be found or will be from
	# a different date.
	#
	$opened = 0;
	$version = 0;
	open(soa, "+<output/$file") && do {
		while (<soa>) {
			chop;
			next unless /^\s+${nowstr}(\d\d)\s*\;\sSerial$/;
			$version = 1 + $1;
		}
		seek(soa, 0, 0);
		$opened = 1;
	};
	if (!$opened) {
		open(soa, ">output/$file") || die "output/$file: $!";
	}

	print soa $DoNotEdit;
	&genSOA('soa', $zone, sprintf("%s%02d", $nowstr, $version));
	print soa "\$INCLUDE sends/$file.data\n";
	truncate(soa, tell(soa));
	close(soa);
}


############################################################################
##
## genSOA
##
## output the actually SOA file statements to the given filehandle.
##

sub genSOA {
	local($fh, $zone, $serial) = @_;
	local($oldfh) = select($fh);
	local($_);

	printf "\$ORIGIN %s.\n", $zone;
	printf "\@\t\tIN SOA\t%s %s (\n",
		$SOA{'masterhost'},
		$SOA{'hostmaster'};
	printf "\t\t\t\t%s\t; Serial\n", $serial;
	foreach ('refresh', 'retry', 'expire', 'minimum') {
		printf "\t\t\t\t%d\t\t; %s\n", $SOA{$_}, $_;
	}
	print "\t\t\t)\n";
	foreach (@NameServers) {
		printf "\t\tIN NS\t%s\n", $_;
	}
	print "\n";
	select($oldfh);
}

############################################################################
##
## produceBootFile
##
## Generate the toplevel boot files for the primary and secondary
## nameservers. (a whole mess of 'primary' or 'secondary' statements).
##

sub produceBootFiles {
	local($zone, $file, $_);
	local(*primary, *secondary);

	open(primary, ">boot.primary.$$") || die "boot.primary: $!";
	open(secondary, ">boot.secondary.$$") || die "boot.secondary: $!";

	$_ =  $DoNotEdit;
	$_ .= "; it is intended to be included from a named.boot file\n";
	print primary $_;
	print secondary $_;

	foreach $zone (sort(keys(%FileByZone))) {
		$file = $FileByZone{$zone};
		printf primary
			"primary\t%-31s\tsends/%s\n",
			$zone, $file;
		printf secondary
			"secondary %-31s %s sends/%s\n",
			$zone, $PriAddr, $file;
	}

	close(primary) || die "close(boot.primary.$$): $!";
	close(secondary) || die "close(boot.secondary.$$): $!";

	&sends'renameOrUnlink("boot.primary.$$", "boot.primary");
	&sends'renameOrUnlink("boot.secondary.$$", "boot.secondary");
}
