#!/usr/bin/perl

# Copyright (c) 1994 by Vixie Enterprises
# 
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies, and that
# the name of Vixie Enterprises not be used in advertising or publicity
# pertaining to distribution of the document or software without specific,
# written prior permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND Vixie Enterprises DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL Vixie Enterprises
# BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
# OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

# $Id: scan-ftp.pl,v 1.11 1995/03/04 01:28:11 brister Exp $

# Credits:
#       Paul Vixie <paul@vix.com> (functional design partner; original coder)
#       Jerry Scharf <scharf@sony.com> (functional design partner)
#       James Brister <brister@vix.com> (major code rework; maintainance)


#
# Scan the home ftp directories of all the groups and move any fully
# transferred file from there into the working area. Then execute other
# scripts to process the data files.
#

# Boilerplate

$sends'Root = $ENV{'SENDS_ROOT'} ;
$sends'Root = "/usr/local/sends" unless defined ($sends'Root) ;
require "$sends'Root/lib/sends-config.pl" ;

#
# Here we go....
#

&getLock ;

#
# Any file that matches the key below gets the corresponding command 
# run on it.
#
%Files = (
	'hosts', "$sends'binDir/do-hosts -w %s -W %s",
	'users', "$sends'binDir/do-users -w %s -W %s"
);


#
# files not matching these keys (or the entries in $Files) will be 
# tagged for removal from the ftp directory. 
#
%OKjunk = (
	'bin', ''
);

@Files = keys(%Files);

#
# files we say should be removed from a group's ftp directory (excluding those
# that match the keys in $OKjunk).
# 
@ShouldRM = ();


# Legal opts: -d and -t 
&parse_options();

#
# Want the config file to take effect if wanted.
#
$Debugging = $sends'Config{'debug'} ;
$Debugging = 0 unless defined $Debugging ;

# Clean out old crap from work directory.
system("rm -f $sends'zoneDir/work/* >/dev/null 2>&1");

do {
	#
	# Array of two keys 'hosts' and 'users'. With the value of
	# each being the oldest file in all the groups' 'work'
	# directories
	#
	%oldestWorkFile = ();

	#
	# iterate over all the group directories and move files from
	# the ftp directory to the 'work' directory'
	#
	&sends'forEachGroup("&main'moveFtpFiles()");

	#
	# Iterate over all the group directories and look for stuff to
	# do in the 'work' directories. Sets up %oldestWorkFile
	#
	&sends'forEachGroup("&main'scanForWork()");

	$somethingHappened = 0;
	foreach $file (keys(%oldestWorkFile)) {
		($path, $age, $domain, $group) = split(/$;/,
						       $oldestWorkFile{$file});
		printf STDERR "oldest(%s) -> %s\n", $file, $path if $Debugging;
		&processWorkFile($file, $path, $age, $domain, $group);
		$somethingHappened++;
	}
} while ($somethingHappened);

if ($#ShouldRM >= $[) {
	print STDERR "Should remove the following junk files:\n";
	foreach (@ShouldRM) {
		print "\t$_\n";
	}
}

exit(0);

################################################################ parse_options
###
### parse_options
###
### parse command line options.
###
### -d == debugging flag
### -t == test (output to stdout not to sendmail).

sub parse_options {
	$opt_d = $opt_t = undef;
	&Getopts('dt');
	if (0) {
		print STDERR "usage: $0 [-d] [-t]\n";
		exit(1);
	}
	$Debugging = $opt_d;
#	if (!$opt_t) {
#		print STDERR "$0: option -t forced on for now\n";
#		$opt_t++;
#	}
}

################################################################ moveFtpFiles
###
### moveFtpFiles
###
### move files from the ftp directory into the working directory.
###
### parameters: PATH to group's top directory. (passed in via a global
###             variable as this routine is eval'ed).
###

sub moveFtpFiles {
	local($path) = $sends'Path;
	local($file, *ftp);

	chdir($path) || die "chdir($path): $!";
	opendir(ftp, 'ftp') || die "opendir($path/ftp): $!";
  file:	while ($file = readdir(ftp)) {
		next file if ($file eq '.') || ($file eq '..');
		if (!defined($Files{$file})) {
			push(@ShouldRM, $path.'/ftp/'.$file)
				unless defined($OKjunk{$file});
			next file;
		}
		next file unless (-f "ftp/$file");
		next file if &sends'isOpen("ftp/$file");
		print STDERR "$path has a ftp/$file\n"		if $Debugging;
		rename("ftp/$file", "work/$file") || die "rename($file): $!";
		if (&sends'isOpen("work/$file")) {
			#
			# race detected: someone opened it between our
			# first isOpen test and our rename.  rename it back.
			#
			rename("work/$file", "ftp/$file") || die "rename: $!";
			next file;
		}
	}
	closedir (ftp) ;
	return 1;
}

################################################################ scanForWork
###
### scanForWork
###
### Look to see if transferred files have been moved into the 'work'
### directory so they can be processed.
###
### Parameters: PATH to group's directory (passed in via a global
###             parameter as this routine is eval'ed).
###

sub scanForWork {
	local($path) = $sends'Path;
	local($file, $age, $old_path, $old_age, $old_domain, $old_group);

	chdir($path) || die "chdir($path): $!";
  file:	foreach $file (@Files) {
		next file unless (-f "work/$file");
		$age = (-M _);
		next file if &sends'isOpen("work/$file");
		print STDERR "$path has a work/$file\n" if $Debugging;

		#
		# we cache the age of the file to be processed so the
                # calling routine can process them in chronological
		# order.
		#
		if (defined($oldestWorkFile{$file})) {
			#
			# this will happen if new data is ftp'd in
			# while the script is running.
			#
			($old_path, $old_age, $old_domain, $old_group) =
				split(/$;/, $oldestWorkFile{$file});
			next file unless $age > $old_age;
		}
		$oldestWorkFile{$file} =
			join($;, "$path/work/$file", $age,
				 $sends'Domain, $sends'Group);
	}
	return 1;
}

############################################################### processWorkFile
###
### processWorkFile
###
### process the files that are in a group's 'work' directory. Output
### in the form of a mail message is generated and sent the DRI of the
### group and to the sends people.
### 
### Parameters: FILE to be processed. PATH to group's main directory.
### AGE of the file (not used). DOMAIN for the GROUP.
###

sub processWorkFile {
	local($file, $path, $age, $domain, $group) = @_;
	local($wd) = &sends'GroupToPath($domain, $group);
	local($_, $dri);
	local(@driList, %control, *notify, *installer);

	chdir($wd) || die "chdir($wd): $!";

	#
	# get necessary information for the sending of the mail message.
	#
	%control = &sends'getControl($domain, $group);
	if (!scalar(%control)) {
		print STDERR "$0: [$domain,$group]: ".$sends'Error."\n";
		return;
	}
	@driList = split(/$;/, $control{'dri'});
	if ($#driList < $[) {
		print STDERR "$0: [$domain,$group]: no DRI list in control\n";
		return;
	}

	if ($opt_t || $sends'Config{'dont_mail'}) {
		open(notify, ">&STDOUT") || die "$0: notify: $!";
	} else {
		open(notify, "|".$sends'Sendmail." -t") || die "$0: notify: $!";
	}

	$_ = "To: ";
	foreach $dri (@driList) {
		printf notify "%s%s", $_, $dri;
		$_ = ",\n    ";
	}
	printf notify "\n";
	printf notify "Cc: %s\n", $sends'NotifyCC;
	printf notify "Subject: new \"%s\" file for [%s, %s] received by SENDS\n",
			$file, $domain, $group;
	printf notify "\n";
	printf notify "-------- installation\n";

	#
	# generate and execute a shell command consisting of the keyed value
	# from %Files, and input coming from the work file itself.
	#
	$_ = sprintf($Files{$file}, $group, $domain);
	$_ .= " -d"						if $Debugging;
	print STDERR "[$_]\n"					if $Debugging;
	open(installer, "$_ 2>&1 |") || die "$0: installer($_): $!";
	while (<installer>) {
		print notify $_;		# capture output and mail it too.
	}
	close(installer);
	printf notify "-------- end of installation\n";
	$_ = $?;
	printf STDERR "Exit Status \t%04x\n", $_ if $Debugging;
	$_ >>= 8;
	if ($_ != 0) {
		printf notify "\nfailed - rejecting input file\n";
		print STDERR "failed - moving $path to bad/$file\n";
		rename("work/$file", "bad/$file") || die "rename($file): $!";
		return;
	}
	printf notify "\nsucceeded - accepting input file\n";
	close(notify) || die "$0: close(notify): $!";
	rename("work/$file", "output/$file") || die "rename(output/$file): $!";
	return;
}
