#!/usr/bin/perl

# Copyright (c) 1994 by Vixie Enterprises
# 
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies, and that
# the name of Vixie Enterprises not be used in advertising or publicity
# pertaining to distribution of the document or software without specific,
# written prior permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND Vixie Enterprises DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL Vixie Enterprises
# BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
# OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

# $Id: sends-config.pl,v 1.10 1995/02/28 16:21:48 brister Exp $

# Credits:
#       Paul Vixie <paul@vix.com> (functional design partner; original coder)
#       Jerry Scharf <scharf@sony.com> (functional design partner)
#       James Brister <brister@vix.com> (major code rework; maintainance)


# Assumptions: the variable $sends'Root will be set to the directory
# that is the root of the sends (source and data) tree

# The %main'CONFIG assoc array must have the keys match the legal
# keywords in the config file.

# make life easier on ourselves.
die "\$sends'Root not defined" unless defined ($sends'Root) ;
unshift (@INC, $sends'Root . "/lib") ;

# check for other versions running
$lockFile = $sends'Root . "/LockFile" ;

sub getLock {
	if ( -e $lockFile ) {
		open (lockFile,"+<$lockFile") || 
			die "$0: Can't open lockFile $!" ;
	} else {
		open (lockFile,"+>$lockFile") || 
			die "$0: Can't open lockFile $!" ;
	}
	$oldfh = select (lockFile) ; $| = 1 ; select ($oldfh) ;
	if ( !flock (lockFile,6) ) {
	        chop ($otherpid = <lockFile>) ;
	        print STDERR "Another process ($otherpid) is running.\n" ;
	        exit (0) ;
	} else {
#		print STDERR "Got lock\n" ;
	}

	print lockFile "$$\n" ;
	truncate (lockFile,tell (lockFile)) ;
}

sub releaseLock {
	if ( !flock (lockFile,8) ) {
		print STDERR "Can't release lock\n" ;
		exit (0) ;
	}
}

##
## a zone mask list entry is of the form:
##	network a.b.c.d netmask e.f.g.h
## where a,b,d,e,f,g,h are numbers.
##
sub checkZoneMaskList {
	local (@l) = split ;

	while ( $#l - $[ + 1 > 0 ) {
		if ( $#l - $[ + 1 < 4 ) { # each entry has 4 fields
			return 0 ;
		}
		if ( $l[0] ne "network" ) {
			return 0 ;
		}
		if ( $l[1] !~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/) {
			return 0 ;
		}
		if ( $l[2] ne "netmask" ) {
			return 0 ;
		}
		if ( $l[3] !~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/) {
			return 0 ;
		}

		shift (@l) ;
		shift (@l) ;
		shift (@l) ;
		shift (@l) ;
	}

	return 1 ;
}

require "sendsConfigLib.pl" ;

package sends ;

$configFileDefault = $ENV{'SENDS_CONFIG'} ;
$configFileDefault = "$sends'Root/data/sends.conf" 
    unless defined $configFileDefault ;

# This array stores all the configuration values that are pulled in from the
# config file. The keywords in the config file are used as keys in this
# array. If a keyword is not found in the config file that is in this array,
# then that is an error. 

@RequiredKeys = ('primary',
		 'name_servers',
		 'hostmaster',
		 'network',
		 'zone_master',
		 'zone_expire',
		 'zone_retry', 
		 'zone_refresh',
		 'zone_minimum',
		 'notify_addr',
		 'bad_domains',
		 'sendmail',
		 'user_key_no_defaults',
		 'user_key_no_dups',
		 'user_key_min_fields',
		 'whois_file'
);

@OptionalKeys = ( 'debug', 'sendmail', 'tmpdir', 'dont_mail', 
			'whois_linelimit', 'named-pid-file' ) ;

%VerifyExprs = ('*', '$config\'Error = "$config\'key is an empty string", 
                      &config\'nonEmptyString',    # catch-all test
	   	'primary', '$config\'Error = "primary is not a dotted quad",
			   /^(\d+)\.(\d+)\.(\d+)\.(\d+)/',
		'zone_expire', "&config'checkPosInt",
#		'zonesize', "&config'checkPosInt(1,4)",
		'zone_retry', "&config'checkPosInt",
		'zone_refresh', "&config'checkPosInt",
		'zone_minimum', "&config'checkPosInt",
		'sendmail', '$config\'Error = "sendmail value is not valid",
			    -f $_ && -x _',
		'debug', "&config'checkPosInt",
		'tmpdir', '$config\'Error = "$config\'key is not a directory", -d ',
		'dont_mail', 1, # '$sends\'Config{\'dont_mail\'} = 1',

);

$flags = ($config'AllowExtras | $config'MustHaveExpr) ;

%defaults = (
	'zone_expire', '604800',
	'zone_retry', '600',
	'zone_refresh', '1800',
	'zone_minimum', '1800',
	'tmpdir', '/var/tmp',
	'debug', '0',
	'dont_mail', '0',
	'whois_file', 'data/whois',
);


# find sendmail in PATH for default.
foreach $dir (split (/:/,$ENV{'PATH'})) {
	if ( -f $dir . "/sendmail" && -x _ ) {
		$default{'sendmail'} = $dir . "/sendmail" ;
		last ;
	}
}

%Config = () ;

die "SENDS config ($configFileDefault) failed: $config'Error\n" 
	unless &config'processConfig (*Config,$configFileDefault,
				      *RequiredKeys, *OptionalKeys,
				      *defaults, *VerifyExprs, $flags) ;

package main ;

require 'getopts.pl' ;
require 'sends-init.pl' ;

1 ;
