#!/usr/bin/perl

# Copyright (c) 1994 by Vixie Enterprises
# 
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies, and that
# the name of Vixie Enterprises not be used in advertising or publicity
# pertaining to distribution of the document or software without specific,
# written prior permission.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND Vixie Enterprises DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL Vixie Enterprises
# BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
# OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
# WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
# ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

# $Id: sends-init.pl,v 1.13 1995/03/04 01:31:40 brister Exp $

# Credits:
#       Paul Vixie <paul@vix.com> (functional design partner; original coder)
#       Jerry Scharf <scharf@sony.com> (functional design partner)
#       James Brister <brister@vix.com> (major code rework; maintainance)



# initialize general purpose (for sends) routines as well as global variables.
# should only really be called by sends-config...

$ENV{'PATH'} .= ':/usr/local/bin';

package sends;

die "\$sends'Root not defined" if ! defined ($sends'Root) ;

##############################################

$topDir = $sends'Root ;		# top of the source and data tree
$dataDir = "$topDir/data";	# top of where the data lives
$outputDir = "$topDir/data/Output";
$binDir = "$topDir/bin";	# standalone SENDS programs live here
$libDir = "$topDir/lib";	# perl etc. libraries live here
$zoneDir = "$topDir/data/Zones"; # dir for the generated DNS files.
$userDir = "$topDir/data/Users"; # dir for the generated USER files.

$Mode = 0664;
$dirMode = 0775;   # not used anywhere, so I'm not putting it in config.
$NotifyCC = $Config{'notify_addr'};	# who gets copied on mail.
$Sendmail = $Config{'sendmail'}; # config.pl sez we have a good value here.
$Debug = $Config{'debug'} ;
$MySep = "\016";

$Error = undef;

umask (0002) ;

##
## cat a file into an array and return it.
##
sub cat {
	local($file) = @_;

	open(file) || do {
		warn "&domain'cat($file): $!";
		return ();
	};
	@file = <file>;
	close(file);
	return @file;
}


##
## Simulate 'mkdir -p'
##
sub mkdir {
	local($path, $mode) = @_;
	local(@path) = split(m:/:, $path);
	local($_);

	$path = '';
	foreach $_ (@path) {
		next unless $_;
		$path .= "/$_";
		next unless (! -d $path);
		mkdir($path, $mode) || return 0;
	}
	return 1;
}

##
## Turn an integer IP address into a a full-on PTR address.
##
sub revAddr {
	local($iaddr, $parts) = @_;
	local($addr) = &fmtAddr($iaddr);
	local(@addr) = split(/\./, $addr);

	$parts = 4 unless defined($parts);
	return join('.', reverse(@addr[0..$parts-1]), 'in-addr.arpa.');
}

##
## Turn a dotted quad address into the integer equivalent.
##
sub iaddrOf {
	local($addr) = @_;
	local($a, $b, $c, $d);

	if ($addr !~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/) {
		return ("$addr is not a dotted quad");
	}
	($a, $b, $c, $d) = ($1, $2, $3, $4);
	if (	$a < 0	||  $a > 255	||
		$b < 0	||  $b > 255	||
		$c < 0	||  $c > 255	||
		$d < 0	||  $d > 255	)
	{
		return ("$addr has a bad octet");
	}
	return ( '', (($a << 24) | ($b << 16) | ($c << 8) | $d) );
}

##
## Turn an integer IP address into a dotted-quad string.
##
sub fmtAddr {
	local($addr) = @_;
	local($_);

	return sprintf("%d.%d.%d.%d",
		0xff & ($addr>>24),
		0xff & ($addr>>16),
		0xff & ($addr>>8),
		0xff & ($addr));
}

## Cache the age of the last modification to the top level data directory
$DomainListAge = -1;
## Cache the list of domains defined inside the top level data directory.
@DomainListCache = ();

##
## return the list of domains we are managing. Creating a new domain is 
## done by just creating the proper subdirectory in the data directory.
##
sub DomainList {
	local(*d);
	local($_, $age);

	#
	# check for new domain this way rather than groping through the
	# whole directory listing each time.
	#
	$age = (-M $dataDir);
	if ($age != $DomainListAge) {
		#
		# must have a new domain in there
		#
		opendir(d, $dataDir) || do {
			$Error = "$dataDir: $!";
			return ();
		};

		#
		# Weed out non-domain directories
		#
		foreach $d ( grep(/^[a-z]/, (readdir(d))) ) {
			push (@DomainListCache,$d) if ( -d "$dataDir/$d" ) ;
		}
		closedir(d);
		$DomainListAge = $age;
	}
	die "ERROR: NO DOMAIN LIST\n" if ( ! @DomainListCache ) ;
	return @DomainListCache;
}


## Cache of the modification times of the domain directories
%GroupListAge = ();
## Cache of the groups inside each domain
%GroupListCache = ();

##
## Scan the data directory for a particular domain and find all the groups
## within that domain (i.e. all the directories listed in there). We also check
## that each group we get has the appropriate directory structure, and a 
## control file.
##
sub GroupList {
	local($domain) = @_;
	local(*d);
	local($_, $age, @grps,@goodgroups,$d);

	# 
	# check the date on the domain's directory. If it hasn't changed
	# then there are no new groups and we can skip reading the
	# directory and just return our cached value.
	#
	$age = (-M "$dataDir/$domain");
	if ($age != $GroupListAge{$domain}) {
		opendir(d, "$dataDir/$domain") || do {
			$Error = "$dataDir/$domain: $!";
			return ();
		};
		$d = "$dataDir/$domain" ;
		@goodgroups = () ;
		@grps = grep(/^[a-z]/, (readdir(d))) ;

		foreach $g ( @grps ) {
			if ( ! -d "$d/$g" ) {
				print STDERR "WARNING: Skipping non-directory $d/$g\n" ;
				next ;
			}
			if ( ! -d "$d/$g/bad" ) {
				print STDERR "WARNING: Skipping group with missing" .
					     " 'bad' directory: $d/$g\n" ;
				next ;
			}
			if ( ! -d "$d/$g/work" ) {
				print STDERR "WARNING: Skipping group with missing" .
					     " 'work' directory: $d/$g\n" ;
				next ;
			}
			if ( ! -d "$d/$g/output" ) {
				print STDERR "WARNING: Skipping group with missing" .
					     " 'output' directory: $d/$g\n" ;
				next ;
			}
			if ( ! -d "$d/$g/ftp" ) {
				print STDERR "WARNING: Skipping group with missing" .
					     " 'ftp' directory: $d/$g\n" ;
				next ;
			}
			if ( ! -f "$d/$g/control" ) {
				print STDERR "WARNING: Skipping group with missing" .
					     " control file: $d/$g\n" ;
				next ;
			}

			push (@goodgroups,$g) ;
		}
			

		# REMIND should we '-d' each returned value?
		$GroupListCache{$domain} = join($;,@goodgroups) ;
		closedir(d);
		$GroupListAge{$domain} = $age;
	}
	return (split(/$;/, $GroupListCache{$domain}));
}


##
## return all the FQDN's (from the control file) of all the groups in all
## the domains.
##
sub fqdnList {
	local($domain, $group, $fqdn);
	local(%ret);

	%ret = ();
	foreach $domain (&DomainList()) {
	    foreach $group (&GroupList($domain)) {
		$fqdn = &fqdnByGroup($domain, $group);
		next unless defined $fqdn;
		$ret{$fqdn} = '';
	    }
	}
	return (keys(%ret));
}


##
## return all the MAIL-DOMAIN's (from the control file) of all the groups
## in all the domains.
##
sub MailDomList {
	local($domain, $group, $maildom);
	local(%ret);

	%ret = ();
	foreach $domain (&DomainList()) {
	    foreach $group (&GroupList($domain)) {
		$maildom = &MailDomByGroup($domain, $group);
		next unless defined $maildom;
		$ret{$maildom} = '';
	    }
	}
	return (keys(%ret));
}


##
## Transform a Group/Domain pair into a directory pathname.
##
sub GroupToPath {
	local($domain, $group) = @_;
	return $dataDir . '/' . $domain . '/' . $group;
}

##
## Parse a sequence of lines formatted in a Mail header style. i.e. each line
## starts with a keyword followed by a colon, followed by some
## indertiminate number of spaces, followed by the value of the header.
## continuation lines are allowed and must start with white space.
##
## Parameter is an open filehandle. Lines are read just until a blank line.
## An associative array is returned, where the keys are the keywords from
## the lines, and values are the rest of the lines.
##
## All key values are tranformed to lower case before used as a key in the
## array. 
##
sub parseHdrs {
	local($fh) = @_;
	local(%ret) = ();
	local($_, $key, $val);

##	print STDERR "parseHdrs($fh)\n";
	$key = '';
	while (<$fh>) {
		chop;
		if (/^$/) {
			last if $key;
			next;
		}
		next if /^\#/;
		if (/^\s+$/) {
			$Error = "$fh:$.: blanks-only line, possible separator";
			return ();
		}
		if (/^\s+(.*)/) {
			next unless $key;	# ignore leading continuations
			$val = $1;
			$ret{$key} .= "\n\t".$val;
		} elsif (/^([\w\-\_]+):\s*(.*)$/) {
			($key, $val) = ($1, $2);
			$key =~ tr/A-Z/a-z/;
			$val =~ s/^\s+//;
			$val =~ s/\s+$//;
			if (defined($ret{$key})) {
				$ret{$key} = join($;, split(/$;/, $ret{$key}),
						      $val);
			} else {
				$ret{$key} = $val;
			}
		} else {
			$Error = "$fh:$.: bad line: \"$_\"";
			return ();
		}
	}
	$Error = undef;
##	print STDERR "parseHdr returning ",scalar(%ret),"\n";
	return %ret;
}


## Cache of ages of control files for each group
%ctrlCacheAge = ();
## cache of the control data for each group.
%ctrlCacheData = ();

##
## Get the control information from the control file in a group's
## directory.
##
## Parameters are the DOMAIN and GROUP to get the control information for.
##
sub getControl {
	local($domain, $group) = @_;
	local($path) = &GroupToPath($domain, $group);
	local(%ret) = ();
	local(%cur);
	local(*fh);
	local($key, $val, $age);

	if (! -f "$path/control") {
		$Error = "$path/control: does not exist";
		return ();
	}

	#
	# check our age cache
	#
	$age = (-M _);
	if (defined($ctrlCacheAge{$domain,$group})) {
		if ($ctrlCacheAge{$domain,$group} == $age) {
##			print STDERR "getControl($domain,$group): CACHE HIT\n";
			%ret = split(/$MySep/, $ctrlCacheData{$domain,$group});
			return %ret;
		}
	}

	# 
	# No cache hit, so pull out the data.
	#
	open(fh, "<$path/control") || return ();
	
	# The parseHdrs routine reads only up to the next blank line, but there
	# could be more than one set; for control files, we glom 'em together
	while (!eof(fh)) {
		%cur = &parseHdrs('fh');
		while (($key,$val) = each(%cur)) {
			$ret{$key} = $val;	# "merge/replace" semantics
		}
	}
	close(fh);
	$Error = '$path/control: empty file' unless scalar(%ret);
	$ctrlCacheAge{$domain,$group} = $age;
	$ctrlCacheData{$domain,$group} = join($MySep, %ret);
	return %ret;
}

##
## Get the FQDN for a group.
##
## Parameter is the DOMAIN and GROUP we want the information for.
##
sub fqdnByGroup {
	local($domain, $group) = @_;
	local(%control) = &getControl($domain, $group);
	local($fqdn);

	return undef unless scalar(%control);
	$fqdn = $control{'fqdn'};
	if (defined($fqdn)) {
		return $fqdn if ($fqdn =~ /\.$domain$/);
		$Error = "fqdnByGroup($domain, $group): $fqdn not in $domain";
		return undef;
	} else {
		return $group .'.'. $domain;
	}
}


##
## Get the MAIL-DOMAIN control info for the DOMAIN and GROUP given as 
## the parameters. The MAIL-DOMAIN is either the value of the control 
## entry of the same name, or (if that doesn't exist) the FQDN for the 
## group.
##
sub MailDomByGroup {
	local($domain, $group) = @_;
	local(%control) = &getControl($domain, $group);
	local($maildom);

	return undef unless scalar(%control);
	$maildom = $control{'mail-domain'};
	if (defined($maildom)) {
		return $maildom if ($maildom =~ /\.$domain$/);
		$Error = "MailDomByGroup($domain,$group): $maildom not in $domain";
		return undef;
	} else {
		return &fqdnByGroup($domain, $group);
	}
}

##
## Munge the semi-regexp NETWORK control information into a slightly more
## usable form.
## 
sub parseNetworks {
	local($_) = @_;
	local($range, $action, $netmask) = split;
	local($error);

	return (undef,
		"bad action \"$action\" in \"$_\": must be accept or deny"
	       ) unless ($action =~ /^(accept|deny)$/);
	return (undef,
		"cannot specify \"deny\" and a netmask")
		if ($action eq "deny" && $netmask) ;
	($range, $error) = &parseNetRange($range);
	return ("$range $action $netmask", "") if defined($range);
	return (undef, $error);
}

##
## Parses the network range values passed in by parseNetworks.
##
## Takes a value looking like: "43.(128-191).(0-255).(0-255)" or
##			       "43.(128:63).(0:255).(0:255)"
##
## and canonicalises it:
##
##	43-43.128-191.0-255.0-255
##
sub parseNetRange {
	local($range) = @_;
	local(@octets) = split(/\./, $range);
	local($_, $literal, $base, $length, $start, $end);

	return (undef, "wrong number of octet specs in \"$range\"")
		unless $#octets == 3;
	foreach (@octets) {
		if (/^(\d+)$/) {
			($literal) = ($1);
			$_ = sprintf("%d-%d", $literal, $literal);
		} elsif (/^\((\d+)\:(\d+)\)$/) {
			($base, $length) = ($1, $2);
			$_ = sprintf("%d-%d", $base, $base + $length);
		} elsif (/^\((\d+)\-(\d+)\)$/) {
			($start, $end) = ($1, $2);
			$_ = sprintf("%d-%d", $start, $end);
		} else {
			return (undef, "bad octet spec \"$_\" in \"$range\"");
		}
	}
	return (join('.', @octets), "");
}


##
## Check that an ADDRESS (in dotted-quad form) is within a RANGE (in the
## form generated by parseNetRange).
##
sub addrInRange {
	local($addr, $range) = @_;
	local(@addr, @start, @end);
	local($_);

	if ($addr !~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/) {
		return (undef, "bad ip address \"$addr\"");
	} else {
		@addr = ($1, $2, $3, $4);
	}

	if ($range !~
	    /^(\d+)\-(\d+)\.(\d+)\-(\d+)\.(\d+)\-(\d+)\.(\d+)\-(\d+)$/
	    ) {
		return (undef, "bad net range \"$range\"");
	} else {
		@start = ($1, $3, $5, $7);
		@end = ($2, $4, $6, $8);
	}

	for ($_ = 0;  $_ < 4;  $_++) {
		return (0, "too low") unless ($addr[$_] >= $start[$_]);
		return (0, "too high") unless ($addr[$_] <= $end[$_]);
	}
	return (1, "ok");
}


##
## Iterate over all the GROUPS in all the DOMAINS and evaluate the
## parameter EXPR for each group.
##
## The variable $sends'Path is set to the Domain/Group pathname for the 
## expression to access.
##
sub forEachGroup {
	local($expr) = @_;
	local($Domain, $Group, $Path);
	local($rval) ;

	foreach $Domain (&DomainList()) {
		print STDERR "// Domain=$Domain\n"		if $Debug;
		foreach $Group (&GroupList($Domain)) {
			print STDERR "// Group=$Group\n"	if $Debug;
			$Path = &GroupToPath($Domain, $Group);
			$rval = eval($expr) ;
			warn "$@\n" if $@ ;
			return unless ($rval > 0);
		}
	}
}

##
## Check if a FILE is open by some other process. 
##
## Uses fstat(1) to do so and if very non-portable.
##
sub isOpen {
	local($file) = @_;
	local(*fstat);
	local($_);

####	USER     CMD          PID   FD MOUNT      INUM MODE         SZ|DV NAME
####	vixie    ftpd       21299    5 /            16 -rwxr-xr-x  541796 /bsd

	open(fstat, "fstat $file |") || return 0 ;  # presume it's not open
	while (<fstat>) {
		chop;
		next if /^USER\s+CMD\s+PID\s+FD/;
		if (/^[\w\_\-\.]\s+[\w\_\-\.]\s+\d+\s+\d+\s+/) {
			close(fstat);
			return 1;
		}
		print STDERR "fstat $file: [$_]\n";
	}
	close(fstat);
	return 0;
}

#sub openNewDB {
#	local($handle, $db, $mode) = @_;
#	local($package) = caller;
#	local($_);
#
#	$db = "$dbDir/$db" unless ($db =~ m:^/:);
#	$mode = $Mode unless ($mode != 0);
#	$DBset{$package."'".$handle} = $db;
#	unlink($db.$$.".db");
#	eval '$_ = dbmopen('.$package."'".$handle.", '".$db.$$."', $mode)";
#	if ($@) {
#		$! = "&domain'openNewDB($handle, $db): eval(dbmopen): $@";
#		return 0;
#	}
#	return $_;
#}

#sub closeNewDB {
#	local($handle, $keep) = @_;
#	local($package) = caller;
#	local($_);
#
#	$db = $DBset{$package."'".$handle};
#	if (!$db) {
#		$! = "&domain'closeNewDB($package"."'"."$handle): not open\n";
#		return 0;
#	}
#	delete $DBset{$package."'".$handle};
#	eval '$_ = dbmclose('.$package."'".$handle.")";
#	if ($@) {
#		$! = "&domain'closeNewDB($handle, $db): eval(dbmclose): $@";
#		return 0;
#	}
#	return $_ unless $_;
#	if ($keep) {
#		rename($db.$$.".db", $db.".db") || do {
#			$! = "rename: $!";
#			return 0;
#		};
#	} else {
#		unlink($db.$$.".db") || do {
#			$! = "unlink: $!";
#			return 0;
#		};
#	}
#	return 1;
#}

##
## Either mv's a FILE to a NEWFILE (if FILE is not the same--cmp-wise--as 
## NEWFILE), or rm's FILE if NEWFILE is the same.
##
sub renameOrUnlink {
	local($from, $to) = @_;

	if (0 == (system("cmp -s $from $to") >> 8)) {
		# no change
		unlink($from) || die "unlink($from): $!";
	} else {
		# some file is missing or changed
		rename($from, $to) || die "rename($from, $to): $!";
	}
}


##
## Do a DNS lookup of the name NAME and records type TYPE using dig(1).
## somewhat non-portable.
##
sub dig {
	local($name, $type) = @_;
	local($_, $ans, $dn, $rr_ttl, $rr_type, $rr_data);
	local(*dig, @ret);

	($namePat = $name) =~ s/\W/\\$&/;
	open(dig, "dig $name $type |") || do {
		$Error = "dig: $!";
		return ();
	};
	while (<dig>) {
		chop;
		if (/^\;\;\s+ANSWERS:$/) {
			$ans++;
			last;
		}
	}
	if (!$ans) {
		$Error = "dig($name, $type): no ANSWER";
		close(dig);
		return ();
	}
	@ret = ();
	$Error = undef;
	while (<dig>) {
		chop;
		last if /^$/;
		($dn, $rr_ttl, $rr_type, $rr_data) = split;
		$dn =~ tr/A-Z/a-z/;
		next unless ($dn =~ /^$namePat/);
		push(@ret, join(' ', $dn, $rr_ttl, $rr_type, $rr_data));
	}
	return @ret;
}

1;
